### code for optimization over t_p (paper https://arxiv.org/abs/2102.08336).
# The full code of the Lindblad simulations is available upon request from Francesco Battistel (battistel.fra@protonmail.com)


def eval_optimization_waveform_params(self, pulse_len, waveform_params, setup):

  waveform_params['pulse_len'] = pulse_len  # NB: pulse_len = t_p-2*t_rise to match the convention in the paper

  # compute time evolution according to Lindbald equation. All system and noise parameters are contained in setup and waveform_params
  superoperator, final_time = self.get_noise_integ_superop(
      waveform_params=waveform_params, setup=setup)

  # returns the population in |20> at the end of T_slot
  cost_func = de.extract_data_optimization(superoperator, final_time, 
                              n_levels_qubit=self.sim_params['n_levels_qubit'],
                              n_levels_readres=self.sim_params['n_levels_readres'],
                              avg_photon_number=setup.qubit_dic['readres']['avg_photon_number'])

  return cost_func

def get_optimized_waveform_params(self, waveform_params, setup):
  # optimize t_p-2*t_rise given all other system parameters being fixed

  # first-order estimate of \tilde{g} as in Eq. (10)
  expected_gtilde = np.abs(2*np.pi * setup.bare_coupling * waveform_params['drive_amp'] * setup.qubit_dic['qubit']['anharm'] /    \
                      (np.sqrt(2) * (setup.qubit_dic['qubit']['oper_freq'] - setup.qubit_dic['readres']['oper_freq']) * \
                      (setup.qubit_dic['qubit']['oper_freq'] - setup.qubit_dic['readres']['oper_freq'] + setup.qubit_dic['qubit']['anharm'])))
  
  kappa = 1/setup.qubit_dic['readres']['t1']
  
  if expected_gtilde > kappa/4:   # if in the underdamped regime
      damped_gtilde = np.sqrt(expected_gtilde**2-kappa**2/16) * np.exp(-kappa/(7*expected_gtilde))
      x0_guess = np.array([np.pi/(2*damped_gtilde),])
      if x0_guess > waveform_params['slot_len']-2*waveform_params['rise_time']:  # if minimum should occur after T_slot, set guess as T_slot-2*t_rise
          x0_guess = np.array([(waveform_params['slot_len']-2*waveform_params['rise_time']),])
  else: # if in the critical or overdamped regime
      x0_guess = np.array([(waveform_params['slot_len']-2*waveform_params['rise_time']),])

  # automatically calls eval_optimization_waveform_params until convergence
  minimization = scipy.optimize.minimize_scalar(self.eval_optimization_waveform_params,
                                            method = 'bounded',
                                            bracket = [max((x0_guess[0]-2*waveform_params['rise_time'])/2, 0),
                                                       x0_guess[0]],  # useful first interval for the search, but not a strict bound
                                            bounds = [0,
                                                       min(1.1 * x0_guess[0], (waveform_params['slot_len']-2*waveform_params['rise_time']))], # actual bounds
                                            tol=1e-8,
                                            args = (waveform_params, setup))

  optimized_pulse_len = minimization.x  # NB: optimized_pulse_len = t_p-2*t_rise to match the convention in the paper

  optimized_waveform_params = waveform_params.copy()
  optimized_waveform_params['pulse_len'] = optimized_pulse_len

  return optimized_waveform_params